@@ -1,30 +1,8 @@
 /*
- * Copyright (c) 2004-2007 Douglas Gilbert.
+ * Copyright (c) 2004-2011 Douglas Gilbert.
  * All rights reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote products
- *    derived from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- *
+ * Use of this source code is governed by a BSD-style
+ * license that can be found in the BSD_LICENSE file.
  */
 
 #include <unistd.h>
@@ -40,16 +18,16 @@
 #endif
 #include "sg_lib.h"
 #include "sg_cmds_basic.h"
-#include "sg_cmds_extra.h"
+#include "sg_cmds_mmc.h"
 
-/* A utility program for the Linux OS SCSI subsystem.
+/* A utility program originally written for the Linux OS SCSI subsystem.
  *
  * This program outputs information provided by a SCSI "Get Configuration"
-   command [0x46] which is only defined for CD/DVDs (in MMC-2,3,4,5).
+   command [0x46] which is only defined for CD/DVDs (in MMC-2,3,4,5,6).
 
 */
 
-static char * version_str = "0.30 20070919";
+static char * version_str = "0.37 20110206";    /* mmc6r02 */
 
 #define MX_ALLOC_LEN 8192
 #define NAME_BUFF_SZ 64
@@ -60,12 +38,14 @@ static char * version_str = "0.30 20070919";
 static unsigned char resp_buffer[MX_ALLOC_LEN];
 
 static struct option long_options[] = {
-        {"brief", 0, 0, '1'},
+        {"brief", 0, 0, 'b'},
         {"current", 0, 0, 'c'},
         {"help", 0, 0, 'h'},
         {"hex", 0, 0, 'H'},
         {"inner-hex", 0, 0, 'i'},
         {"list", 0, 0, 'l'},
+        {"raw", 0, 0, 'R'},
+        {"readonly", 0, 0, 'q'},
         {"rt", 1, 0, 'r'},
         {"starting", 1, 0, 's'},
         {"verbose", 0, 0, 'v'},
@@ -73,15 +53,17 @@ static struct option long_options[] = {
         {0, 0, 0, 0},
 };
 
-static void usage()
+
+static void
+usage()
 {
     fprintf(stderr,
             "Usage:  sg_get_config [--brief] [--current] [--help] [--hex] "
             "[--inner-hex]\n"
-            "                      [--list] [--rt=RT] [--starting=FC] "
-            "[--verbose]\n"
-            "                      [--version] DEVICE\n"
-            "  where\n"
+            "                      [--list] [--raw] [--readonly] [--rt=RT]\n"
+            "                      [--starting=FC] [--verbose] [--version] "
+            "DEVICE\n"
+            "  where:\n"
             "    --brief|-b       only give feature names of DEVICE "
             "(don't decode)\n"
             "    --current|-c     equivalent to '--rt=1' (show "
@@ -92,6 +74,9 @@ static void usage()
             "features in hex\n"
             "    --list|-l        list all known features + profiles "
             "(ignore DEVICE)\n"
+            "    --raw|-R         output in binary (to stdout)\n"
+            "    --readonly|-q    open DEVICE read-only (def: open it "
+            "read-write)\n"
             "    --rt=RT|-r RT    default value is 0\n"
             "                     0 -> all feature descriptors (regardless "
             "of currency)\n"
@@ -105,14 +90,14 @@ static void usage()
             "Get configuration information for MMC drive and/or media\n");
 }
 
-struct code_desc {
-        int code;
+struct val_desc_t {
+        int val;
         const char * desc;
 };
 
-static struct code_desc profile_desc_arr[] = {
+static struct val_desc_t profile_desc_arr[] = {
         {0x0, "No current profile"},
-        {0x1, "Non-removable disk"},
+        {0x1, "Non-removable disk (obs)"},
         {0x2, "Removable disk"},
         {0x3, "Magneto optical erasable"},
         {0x4, "Optical write once"},
@@ -127,6 +112,8 @@ static struct code_desc profile_desc_arr[] = {
         {0x14, "DVD-RW sequential recording"},
         {0x15, "DVD-R dual layer sequental recording"},
         {0x16, "DVD-R dual layer jump recording"},
+        {0x17, "DVD-RW dual layer"},
+        {0x18, "DVD-Download disc recording"},
         {0x1a, "DVD+RW"},
         {0x1b, "DVD+R"},
         {0x20, "DDCD-ROM"},
@@ -141,17 +128,21 @@ static struct code_desc profile_desc_arr[] = {
         {0x50, "HD DVD-ROM"},
         {0x51, "HD DVD-R"},
         {0x52, "HD DVD-RAM"},
+        {0x53, "HD DVD-RW"},
+        {0x58, "HD DVD-R dual layer"},
+        {0x5a, "HD DVD-RW dual layer"},
         {0xffff, "Non-conforming profile"},
+        {-1, NULL},
 };
 
-static const char * get_profile_str(int profile_num, char * buff)
+static const char *
+get_profile_str(int profile_num, char * buff)
 {
-    int k, num;
+    const struct val_desc_t * pdp;
 
-    num = sizeof(profile_desc_arr) / sizeof(profile_desc_arr[0]);
-    for (k = 0; k < num; ++k) {
-        if (profile_desc_arr[k].code == profile_num) {
-            strcpy(buff, profile_desc_arr[k].desc);
+    for (pdp = profile_desc_arr; pdp->desc; ++pdp) {
+        if (pdp->val == profile_num) {
+            strcpy(buff, pdp->desc);
             return buff;
         }
     }
@@ -159,7 +150,7 @@ static const char * get_profile_str(int profile_num, char * buff)
     return buff;
 }
 
-static struct code_desc feature_desc_arr[] = {
+static struct val_desc_t feature_desc_arr[] = {
         {0x0, "Profile list"},
         {0x1, "Core"},
         {0x2, "Morphing"},
@@ -189,6 +180,8 @@ static struct code_desc feature_desc_arr[] = {
         {0x31, "Double density CD-R write"},
         {0x32, "Double density CD-RW write"},
         {0x33, "Layer jump recording"},
+        {0x34, "LJ rigid restricted oberwrite"},
+        {0x35, "Stop long operation"},
         {0x37, "CD-RW media write support"},
         {0x38, "BD-R POW"},
         {0x3a, "DVD+RW dual layer"},
@@ -198,6 +191,7 @@ static struct code_desc feature_desc_arr[] = {
         {0x42, "TSR (timely safe recording)"},
         {0x50, "HD DVD read"},
         {0x51, "HD DVD write"},
+        {0x52, "HD DVD-RW fragment recording"},
         {0x80, "Hybrid disc"},
         {0x100, "Power management"},
         {0x101, "SMART"},
@@ -213,17 +207,21 @@ static struct code_desc feature_desc_arr[] = {
         {0x10b, "DVD CPRM"},
         {0x10c, "Firmware information"},
         {0x10d, "AACS"},
+        {0x10e, "DVD CSS managed recording"},
         {0x110, "VCPS"},
+        {0x113, "SecurDisc"},
         {0x120, "BD CPS"},
+        {0x142, "OSSC"},
 };
 
-static const char * get_feature_str(int feature_num, char * buff)
+static const char *
+get_feature_str(int feature_num, char * buff)
 {
     int k, num;
 
     num = sizeof(feature_desc_arr) / sizeof(feature_desc_arr[0]);
     for (k = 0; k < num; ++k) {
-        if (feature_desc_arr[k].code == feature_num) {
+        if (feature_desc_arr[k].val == feature_num) {
             strcpy(buff, feature_desc_arr[k].desc);
             return buff;
         }
@@ -232,7 +230,17 @@ static const char * get_feature_str(int feature_num, char * buff)
     return buff;
 }
 
-static void decode_feature(int feature, unsigned char * ucp, int len)
+static void
+dStrRaw(const char * str, int len)
+{
+    int k;
+
+    for (k = 0 ; k < len; ++k)
+        printf("%c", str[k]);
+}
+
+static void
+decode_feature(int feature, unsigned char * ucp, int len)
 {
     int k, num, n, profile;
     char buff[128];
@@ -269,7 +277,7 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         case 4: cp = "IEEE 1394A"; break;
         case 5: cp = "Fibre channel"; break;
         case 6: cp = "IEEE 1394B"; break;
-        case 7: cp = "serial ATAPI"; break;
+        case 7: cp = "Serial ATAPI"; break;
         case 8: cp = "USB (both 1 and 2)"; break;
         case 0xffff: cp = "vendor unique"; break;
         default:
@@ -316,8 +324,9 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
             break;
         }
         printf("      Loading mechanism: %s\n", cp);
-        printf("      Eject=%d, Prevent jumper=%d, Lock=%d\n",
-               !!(ucp[4] & 0x8), !!(ucp[4] & 0x4), !!(ucp[4] & 0x1));
+        printf("      Load=%d, Eject=%d, Prevent jumper=%d, Lock=%d\n",
+               !!(ucp[4] & 0x10), !!(ucp[4] & 0x8), !!(ucp[4] & 0x4),
+               !!(ucp[4] & 0x1));
         break;
     case 4:     /* Write protect */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
@@ -346,11 +355,13 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
     case 0x22:     /* Sector erasable */
     case 0x26:     /* Restricted overwrite */
     case 0x27:     /* CDRW CAV write */
+    case 0x35:     /* Stop long operation */
     case 0x38:     /* BD-R pseudo-overwrite (POW) */
     case 0x42:     /* TSR (timely safe recording) */
     case 0x100:    /* Power management */
     case 0x109:    /* Media serial number */
     case 0x110:    /* VCPS */
+    case 0x113:    /* SecurDisc */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
                feature);
@@ -370,12 +381,9 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
                feature);
-        if (len < 8) {
-            printf("      additional length [%d] too short\n", len - 4);
-            break;
-        }
-        printf("      MULTI110=%d, Dual-R=%d\n", !!(ucp[4] & 0x1),
-               !!(ucp[6] & 0x1));
+        if (len > 7)
+            printf("      MULTI110=%d, Dual-RW=%d, Dual-R=%d\n",
+                   !!(ucp[4] & 0x1), !!(ucp[6] & 0x2), !!(ucp[6] & 0x1));
         break;
     case 0x20:     /* Random writable */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
@@ -407,14 +415,15 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         for (k = 0; k < num; ++k)
             printf("        %d\n", ucp[8 + k]);
         break;
+    /* case 0x22:     Sector erasable -> see 0x1d entry */
     case 0x23:     /* Formattable */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
                feature);
         if (len > 4)
-            printf("      BD-RE: RENoSA=%d, Expand=%d, QCert=%d, Cert=%d\n",
-                   !!(ucp[4] & 0x8), !!(ucp[4] & 0x4), !!(ucp[4] & 0x2),
-                   !!(ucp[4] & 0x1));
+            printf("      BD-RE: RENoSA=%d, Expand=%d, QCert=%d, Cert=%d, "
+                   "FRF=%d\n", !!(ucp[4] & 0x8), !!(ucp[4] & 0x4),
+                   !!(ucp[4] & 0x2), !!(ucp[4] & 0x1), !!(ucp[5] & 0x80));
         if (len > 8)
             printf("      BD-R: RRM=%d\n", !!(ucp[8] & 0x1));
         break;
@@ -437,6 +446,8 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         printf("      Logical block size=0x%x, blocking=0x%x, PP=%d\n",
                num, ((ucp[8] << 8) + ucp[9]), !!(ucp[10] & 0x1));
         break;
+    /* case 0x26:     Restricted overwrite -> see 0x1d entry */
+    /* case 0x27:     CDRW CAV write -> see 0x1d entry */
     case 0x28:     /* MRW  (Mount Rainier reWriteable) */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
@@ -528,7 +539,7 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
             printf("      additional length [%d] too short\n", len - 4);
             break;
         }
-        printf("      BUF=%d, RDL=%d, Test write=%d, DVD-RW=%d\n",
+        printf("      BUF=%d, RDL=%d, Test write=%d, DVD-RW SL=%d\n",
                !!(ucp[4] & 0x40), !!(ucp[4] & 0x8), !!(ucp[4] & 0x4),
                !!(ucp[4] & 0x2));
         break;
@@ -545,6 +556,18 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         for (k = 0; k < num; ++k)
             printf("        %d\n", ucp[8 + k]);
         break;
+    case 0x34:     /* Layer jump rigid restricted overwrite */
+        printf("    version=%d, persist=%d, current=%d [0x%x]\n",
+               ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
+               feature);
+        if (len < 8) {
+            printf("      additional length [%d] too short\n", len - 4);
+            break;
+        }
+        printf("      CLJB=%d\n", !!(ucp[4] & 0x1));
+        printf("      Buffer block size=%d\n", ucp[7]);
+        break;
+    /* case 0x35:     Stop long operation -> see 0x1d entry */
     case 0x37:     /* CD-RW media write support */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
@@ -555,6 +578,7 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         }
         printf("      CD-RW media sub-type support (bitmask)=0x%x\n", ucp[5]);
         break;
+    /* case 0x38:     BD-R pseudo-overwrite (POW) -> see 0x1d entry */
     case 0x3a:     /* DVD+RW dual layer */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
@@ -631,6 +655,7 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
                (ucp[28] << 8) + ucp[29],
                (ucp[30] << 8) + ucp[31]);
         break;
+    /* case 0x42:     TSR (timely safe recording) -> see 0x1d entry */
     case 0x50:     /* HD DVD Read */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
@@ -653,6 +678,16 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         printf("      HD DVD-R=%d, HD DVD-RAM=%d\n", !!(ucp[4] & 0x1),
                !!(ucp[6] & 0x1));
         break;
+    case 0x52:     /* HD DVD-RW fragment recording */
+        printf("    version=%d, persist=%d, current=%d [0x%x]\n",
+               ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
+               feature);
+        if (len < 8) {
+            printf("      additional length [%d] too short\n", len - 4);
+            break;
+        }
+        printf("      BGP=%d\n", !!(ucp[4] & 0x1));
+        break;
     case 0x80:     /* Hybrid disc */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
@@ -663,6 +698,7 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         }
         printf("      RI=%d\n", !!(ucp[4] & 0x1));
         break;
+    /* case 0x100:    Power management -> see 0x1d entry */
     case 0x101:    /* SMART */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
@@ -684,7 +720,7 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         printf("      SCC=%d, SDP=%d, highest slot number=%d\n",
                !!(ucp[4] & 0x10), !!(ucp[4] & 0x4), (ucp[7] & 0x1f));
         break;
-    case 0x103:    /* CD audio external play */
+    case 0x103:    /* CD audio external play (obsolete) */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
                feature);
@@ -700,11 +736,12 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
                feature);
-        if (len < 8) {
+        if (len < 4) {
             printf("      additional length [%d] too short\n", len - 4);
             break;
         }
-        printf("      M5=%d\n", !!(ucp[4] & 0x1));
+        if (len > 4)
+            printf("      M5=%d\n", !!(ucp[4] & 0x1));
         break;
     case 0x105:    /* Timeout */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
@@ -748,14 +785,15 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         buff[n] = '\0';
         printf("      Drive serial number: %s\n", buff);
         break;
+    /* case 0x109:    Media serial number -> see 0x1d entry */
     case 0x10a:    /* Disc control blocks */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
                feature);
         printf("      Disc control blocks:\n");
         for (k = 4; k < len; k += 4) {
-            printf("        0x%x\n", (ucp[k] << 24) + (ucp[k + 1] << 16) +
-                   (ucp[k + 2] << 8) + ucp[k + 3]);
+            printf("        0x%x\n", ((unsigned int)ucp[k] << 24) +
+                   (ucp[k + 1] << 16) + (ucp[k + 2] << 8) + ucp[k + 3]);
         }
         break;
     case 0x10b:    /* DVD CPRM */
@@ -792,6 +830,19 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
         printf("      Number of AGIDs=%d, AACS version=%d\n",
                (ucp[6] & 0xf), ucp[7]);
         break;
+    case 0x10e:    /* DVD CSS managed recording */
+        printf("    version=%d, persist=%d, current=%d [0x%x]\n",
+               ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
+               feature);
+        if (len < 8) {
+            printf("      additional length [%d] too short\n", len - 4);
+            break;
+        }
+        printf("      Maximum number of scrambled extent information "
+               "entries=%d\n", ucp[4]);
+        break;
+    /* case 0x110:    VCPS -> see 0x1d entry */
+    /* case 0x113:    SecurDisc -> see 0x1d entry */
     case 0x120:    /* BD CPS */
         printf("    version=%d, persist=%d, current=%d [0x%x]\n",
                ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
@@ -804,6 +855,23 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
                "SACs=%d\n", ((ucp[5] >> 4) & 0xf), (ucp[5] & 0xf),
                ucp[6] & 0x3);
         break;
+    case 0x142:    /* OSSC (Optical Security Subsystem Class) */
+        printf("    version=%d, persist=%d, current=%d [0x%x]\n",
+               ((ucp[2] >> 2) & 0xf), !!(ucp[2] & 0x2), !!(ucp[2] & 0x1),
+               feature);
+        if (len < 8) {
+            printf("      additional length [%d] too short\n", len - 4);
+            break;
+        }
+        printf("    PSAU=%d, LOSPB=%d, ME=%d\n", !!(ucp[4] & 0x80),
+               !!(ucp[4] & 0x40), !!(ucp[4] & 0x1));
+        num = ucp[5];
+        printf("      Profile numbers:\n");
+        for (k = 6; (num > 0) && (k < len); --num, k += 2) {
+            printf("        %d\n",
+                   ((unsigned int)ucp[k] << 8) + ucp[k + 1]);
+        }
+        break;
     default:
         printf("    Unknown feature [0x%x], version=%d persist=%d, "
                "current=%d\n", feature, ((ucp[2] >> 2) & 0xf),
@@ -813,49 +881,51 @@ static void decode_feature(int feature, unsigned char * ucp, int len)
     }
 }
 
-static void decode_config(unsigned char * resp, int max_resp_len, int len,
-                          int brief, int inner_hex)
+static void
+decode_config(unsigned char * resp, int max_resp_len, int len, int brief,
+              int inner_hex)
 {
-    int k, curr_profile, extra, feature;
+    int k, curr_profile, extra_len, feature;
     unsigned char * ucp;
     char buff[128];
 
     if (max_resp_len < len) {
-        printf("<<<warning: response to long for buffer, resp_len=%d>>>\n",
-               len);
+        fprintf(stderr, "<<<warning: response to long for buffer, "
+                "resp_len=%d>>>\n", len);
             len = max_resp_len;
     }
     if (len < 8) {
-        printf("response length too short: %d\n", len);
+        fprintf(stderr, "response length too short: %d\n", len);
         return;
     }
     curr_profile = (resp[6] << 8) + resp[7];
     if (0 == curr_profile)
-        printf("No current profile\n");
+        fprintf(stderr, "No current profile\n");
     else
         printf("Current profile: %s\n", get_profile_str(curr_profile, buff));
     printf("Features%s:\n", (brief ? " (in brief)" : ""));
     ucp = resp + 8;
     len -= 8;
-    for (k = 0; k < len; k += extra, ucp += extra) {
-        extra = 4 + ucp[3];
+    for (k = 0; k < len; k += extra_len, ucp += extra_len) {
+        extra_len = 4 + ucp[3];
         feature = (ucp[0] << 8) + ucp[1];
         printf("  %s feature\n", get_feature_str(feature, buff));
         if (brief)
             continue;
         if (inner_hex) {
-            dStrHex((const char *)ucp, extra, 1);
+            dStrHex((const char *)ucp, extra_len, 1);
             continue;
         }
-        if (0 != (extra % 4))
-        printf("    additional length [%d] not a multiple of 4, ignore\n",
-               extra - 4);
+        if (0 != (extra_len % 4))
+            printf("    additional length [%d] not a multiple of 4, ignore\n",
+                   extra_len - 4);
         else
-            decode_feature(feature, ucp, extra);
+            decode_feature(feature, ucp, extra_len);
     }
 }
 
-static void list_known(int brief)
+static void
+list_known(int brief)
 {
     int k, num;
 
@@ -863,25 +933,28 @@ static void list_known(int brief)
     printf("Known features:\n");
     for (k = 0; k < num; ++k)
         printf("  %s [0x%x]\n", feature_desc_arr[k].desc,
-               feature_desc_arr[k].code);
+               feature_desc_arr[k].val);
     if (! brief) {
         printf("Known profiles:\n");
         num = sizeof(profile_desc_arr) / sizeof(profile_desc_arr[0]);
         for (k = 0; k < num; ++k)
             printf("  %s [0x%x]\n", profile_desc_arr[k].desc,
-                   profile_desc_arr[k].code);
+                   profile_desc_arr[k].val);
     }
 }
 
 
-int main(int argc, char * argv[])
+int
+main(int argc, char * argv[])
 {
     int sg_fd, res, c, len;
     int peri_type = 0;
     int brief = 0;
-    int hex = 0;
+    int do_hex = 0;
     int inner_hex = 0;
     int list = 0;
+    int do_raw = 0;
+    int readonly = 0;
     int rt = 0;
     int starting = 0;
     int verbose = 0;
@@ -894,7 +967,7 @@ int main(int argc, char * argv[])
     while (1) {
         int option_index = 0;
 
-        c = getopt_long(argc, argv, "bchHilr:s:vV", long_options,
+        c = getopt_long(argc, argv, "bchHilqr:Rs:vV", long_options,
                         &option_index);
         if (c == -1)
             break;
@@ -911,7 +984,7 @@ int main(int argc, char * argv[])
             usage();
             return 0;
         case 'H':
-            hex = 1;
+            ++do_hex;
             break;
         case 'i':
             inner_hex = 1;
@@ -919,6 +992,9 @@ int main(int argc, char * argv[])
         case 'l':
             list = 1;
             break;
+        case 'q':
+            ++readonly;
+            break;
         case 'r':
             rt = sg_get_num(optarg);
             if ((rt < 0) || (rt > 3)) {
@@ -926,6 +1002,9 @@ int main(int argc, char * argv[])
                 return SG_LIB_SYNTAX_ERROR;
             }
             break;
+        case 'R':
+            ++do_raw;
+            break;
         case 's':
             starting = sg_get_num(optarg);
             if ((starting < 0) || (starting > 0xffff)) {
@@ -974,14 +1053,17 @@ int main(int argc, char * argv[])
         return SG_LIB_FILE_ERROR;
     }
     if (0 == sg_simple_inquiry(sg_fd, &inq_resp, 1, verbose)) {
-        printf("  %.8s  %.16s  %.4s\n", inq_resp.vendor, inq_resp.product,
-               inq_resp.revision);
+        if (0 == do_raw)
+            printf("  %.8s  %.16s  %.4s\n", inq_resp.vendor, inq_resp.product,
+                   inq_resp.revision);
         peri_type = inq_resp.peripheral_type;
         cp = sg_get_pdt_str(peri_type, sizeof(buff), buff);
-        if (strlen(cp) > 0)
-            printf("  Peripheral device type: %s\n", cp);
-        else
-            printf("  Peripheral device type: 0x%x\n", peri_type);
+        if (0 == do_raw) {
+            if (strlen(cp) > 0)
+                printf("  Peripheral device type: %s\n", cp);
+            else
+                printf("  Peripheral device type: 0x%x\n", peri_type);
+        }
     } else {
         fprintf(stderr, ME "%s doesn't respond to a SCSI INQUIRY\n",
                 device_name);
@@ -989,23 +1071,31 @@ int main(int argc, char * argv[])
     }
     sg_cmds_close_device(sg_fd);
 
-    sg_fd = sg_cmds_open_device(device_name, 0 /* rw */, verbose);
+    sg_fd = sg_cmds_open_device(device_name, readonly, verbose);
     if (sg_fd < 0) {
         fprintf(stderr, ME "open error (rw): %s\n", safe_strerror(-sg_fd));
         return SG_LIB_FILE_ERROR;
     }
+    if (do_raw) {
+        if (sg_set_binary_mode(STDOUT_FILENO) < 0) {
+            perror("sg_set_binary_mode");
+            return SG_LIB_FILE_ERROR;
+        }
+    }
 
-    res = sg_ll_get_config(sg_fd, rt, starting, resp_buffer, 
+    res = sg_ll_get_config(sg_fd, rt, starting, resp_buffer,
                               sizeof(resp_buffer), 1, verbose);
     ret = res;
     if (0 == res) {
         len = (resp_buffer[0] << 24) + (resp_buffer[1] << 16) +
               (resp_buffer[2] << 8) + resp_buffer[3] + 4;
-        if (hex) {
+        if (do_hex) {
             if (len > (int)sizeof(resp_buffer))
                 len = sizeof(resp_buffer);
             dStrHex((const char *)resp_buffer, len, 0);
-        } else
+        } else if (do_raw)
+            dStrRaw((const char *)resp_buffer, len);
+        else
             decode_config(resp_buffer, sizeof(resp_buffer), len, brief,
                           inner_hex);
     } else if (SG_LIB_CAT_INVALID_OP == res)
