@@ -1,3 +1,23 @@
+/* A utility program for the Linux OS SCSI generic ("sg") device driver.
+*  Copyright (C) 2001 - 2012 D. Gilbert
+*  This program is free software; you can redistribute it and/or modify
+*  it under the terms of the GNU General Public License as published by
+*  the Free Software Foundation; either version 2, or (at your option)
+*  any later version.
+
+   This program reads data from the given SCSI device (typically a disk
+   or cdrom) and discards that data. Its primary goal is to time
+   multiple reads all starting from the same logical address. Its interface
+   is a subset of another member of this package: sg_dd which is a
+   "dd" variant. The input file can be a scsi generic device, a block device,
+   a raw device or a seekable file. Streams such as stdin are not acceptable.
+   The block size ('bs') is assumed to be 512 if not given.
+
+   This version should compile with Linux sg drivers with version numbers
+   >= 30000 . For mmap-ed IO the sg version number >= 30122 .
+
+*/
+
 #define _XOPEN_SOURCE 500
 #ifndef _GNU_SOURCE
 #define _GNU_SOURCE
@@ -11,13 +31,15 @@
 #include <signal.h>
 #include <ctype.h>
 #include <errno.h>
+#define __STDC_FORMAT_MACROS 1
+#include <inttypes.h>
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/sysmacros.h>
 #include <sys/mman.h>
 #include <sys/time.h>
-#include <linux/major.h> 
+#include <linux/major.h>
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -25,27 +47,8 @@
 #include "sg_lib.h"
 #include "sg_io_linux.h"
 
-/* A utility program for the Linux OS SCSI generic ("sg") device driver.
-*  Copyright (C) 2001 - 2007 D. Gilbert
-*  This program is free software; you can redistribute it and/or modify
-*  it under the terms of the GNU General Public License as published by
-*  the Free Software Foundation; either version 2, or (at your option)
-*  any later version.
 
-   This program reads data from the given SCSI device (typically a disk
-   or cdrom) and discards that data. Its primary goal is to time
-   multiple reads all starting from the same logical address. Its interface
-   is a subset of another member of this package: sg_dd which is a 
-   "dd" variant. The input file can be a scsi generic device, a block device,
-   a raw device or a seekable file. Streams such as stdin are not acceptable.
-   The block size ('bs') is assumed to be 512 if not given. 
-
-   This version should compile with Linux sg drivers with version numbers
-   >= 30000 . For mmap-ed IO the sg version number >= 30122 .
-
-*/
-
-static const char * version_str = "1.17 20070714";
+static const char * version_str = "1.19 20121211";
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_BLOCKS_PER_TRANSFER 128
@@ -63,7 +66,7 @@ static const char * version_str = "1.17 20070714";
 
 #ifndef RAW_MAJOR
 #define RAW_MAJOR 255   /*unlikey value */
-#endif 
+#endif
 
 #define FT_OTHER 1              /* filetype other than sg or raw device */
 #define FT_SG 2                 /* filetype is sg char device */
@@ -75,9 +78,9 @@ static const char * version_str = "1.17 20070714";
 
 static int sum_of_resids = 0;
 
-static long long dd_count = -1;
-static long long orig_count = 0;
-static long long in_full = 0;
+static int64_t dd_count = -1;
+static int64_t orig_count = 0;
+static int64_t in_full = 0;
 static int in_partial = 0;
 
 static int pack_id_count = 0;
@@ -102,8 +105,8 @@ static void print_stats(int iters, const char * str)
 {
     if (orig_count > 0) {
         if (0 != dd_count)
-            fprintf(stderr, "  remaining block count=%lld\n", dd_count);
-        fprintf(stderr, "%lld+%d records in", in_full - in_partial,
+            fprintf(stderr, "  remaining block count=%"PRId64"\n", dd_count);
+        fprintf(stderr, "%"PRId64"+%d records in", in_full - in_partial,
                 in_partial);
         if (iters > 0)
             fprintf(stderr, ", %s commands issued: %d\n", (str ? str : ""),
@@ -187,7 +190,8 @@ static void usage()
            "    no_dxfer 1->DMA to kernel buffers only, not user space, "
            "0->normal(def)\n"
            "    odir     1->open block device O_DIRECT, 0->don't (def)\n"
-           "    skip     each transfer starts at this logical address (def=0)\n"
+           "    skip     each transfer starts at this logical address "
+           "(def=0)\n"
            "    time     0->do nothing(def), 1->time from 1st cmd, 2->time "
            "from 2nd, ...\n"
            "    verbose  increase level of verbosity (def: 0)\n"
@@ -198,7 +202,7 @@ static void usage()
 }
 
 static int sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz,
-                             unsigned int blocks, long long start_block,
+                             unsigned int blocks, int64_t start_block,
                              int write_true, int fua, int dpo)
 {
     int rd_opcode[] = {0x8, 0x28, 0xa8, 0x88};
@@ -293,7 +297,7 @@ static int sg_build_scsi_cdb(unsigned char * cdbp, int cdb_sz,
    1 -> recoverable (ENOMEM), 2 -> try again (e.g. unit attention),
    3 -> try again (e.g. aborted command), -1 -> other unrecoverable error */
 static int sg_bread(int sg_fd, unsigned char * buff, int blocks,
-                    long long from_block, int bs, int cdbsz,
+                    int64_t from_block, int bs, int cdbsz,
                     int fua, int dpo, int * diop, int do_mmap,
                     int no_dxfer)
 {
@@ -303,7 +307,7 @@ static int sg_bread(int sg_fd, unsigned char * buff, int blocks,
     struct sg_io_hdr io_hdr;
 
     if (sg_build_scsi_cdb(rdCmd, cdbsz, blocks, from_block, 0, fua, dpo)) {
-        fprintf(stderr, ME "bad cdb build, from_block=%lld, blocks=%d\n",
+        fprintf(stderr, ME "bad cdb build, from_block=%"PRId64", blocks=%d\n",
                 from_block, blocks);
         return -1;
     }
@@ -314,7 +318,8 @@ static int sg_bread(int sg_fd, unsigned char * buff, int blocks,
     if (blocks > 0) {
         io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
         io_hdr.dxfer_len = bs * blocks;
-        if (! do_mmap) /* not required: shows dxferp unused during mmap-ed IO */
+        /* next: shows dxferp unused during mmap-ed IO */
+        if (! do_mmap)
             io_hdr.dxferp = buff;
         if (diop && *diop)
             io_hdr.flags |= SG_FLAG_DIRECT_IO;
@@ -372,7 +377,7 @@ static int sg_bread(int sg_fd, unsigned char * buff, int blocks,
         return -1;
     }
     if (blocks > 0) {
-        if (diop && *diop && 
+        if (diop && *diop &&
             ((io_hdr.info & SG_INFO_DIRECT_IO_MASK) != SG_INFO_DIRECT_IO))
             *diop = 0;      /* flag that dio not done (completely) */
         sum_of_resids += io_hdr.resid;
@@ -387,7 +392,7 @@ static int sg_bread(int sg_fd, unsigned char * buff, int blocks,
 
 int main(int argc, char * argv[])
 {
-    long long skip = 0;
+    int64_t skip = 0;
     int bs = 0;
     int bpt = DEF_BLOCKS_PER_TRANSFER;
     char str[STR_SZ];
@@ -582,7 +587,7 @@ int main(int argc, char * argv[])
             if (verbose > 2) {
                 if (ioctl(infd, SG_GET_RESERVED_SIZE, &t) >= 0)
                     fprintf(stderr, "  SG_GET_RESERVED_SIZE yields: %d\n", t);
-            } 
+            }
             t = bs * bpt;
             if ((do_mmap) && (0 != (t % psz)))
                 t = ((t / psz) + 1) * psz;    /* round up to next pagesize */
@@ -672,7 +677,7 @@ int main(int argc, char * argv[])
     start_tm.tv_usec = 0;
 
     if (verbose && (dd_count < 0))
-        fprintf(stderr, "About to issue %lld zero block SCSI READs\n",
+        fprintf(stderr, "About to issue %"PRId64" zero block SCSI READs\n",
                 0 - dd_count);
 
     /* main loop */
@@ -696,12 +701,12 @@ int main(int argc, char * argv[])
                     buf_sz = MIN_RESERVED_SIZE;
                 blocks_per = (buf_sz + bs - 1) / bs;
                 blocks = blocks_per;
-                fprintf(stderr, 
+                fprintf(stderr,
                         "Reducing read to %d blocks per loop\n", blocks_per);
                 res = sg_bread(infd, wrkPos, blocks, skip, bs, scsi_cdbsz,
                                fua, dpo, &dio_tmp, do_mmap, no_dxfer);
             } else if (2 == res) {
-                fprintf(stderr, 
+                fprintf(stderr,
                         "Unit attention, try again (r)\n");
                 res = sg_bread(infd, wrkPos, blocks, skip, bs, scsi_cdbsz,
                                fua, dpo, &dio_tmp, do_mmap, no_dxfer);
@@ -749,7 +754,7 @@ int main(int argc, char * argv[])
                    (EINTR == errno))
                 ;
             if (res < 0) {
-                snprintf(ebuff, EBUFF_SZ, ME "reading, skip=%lld ", skip);
+                snprintf(ebuff, EBUFF_SZ, ME "reading, skip=%"PRId64" ", skip);
                 perror(ebuff);
                 break;
             } else if (res < blocks * bs) {
@@ -796,7 +801,7 @@ int main(int argc, char * argv[])
                 b = 0.0;
                 c = 0.0;
             }
-    
+
             if (1 == do_time) {
                 fprintf(stderr, "Time for all %s commands was "
                         "%d.%06d secs", read_str, (int)res_tm.tv_sec,
@@ -813,7 +818,7 @@ int main(int argc, char * argv[])
                     fprintf(stderr, ", %.2f MB/sec\n", c / (a * 1000000.0));
                 else
                     fprintf(stderr, "\n");
-            } else { 
+            } else {
                 fprintf(stderr, "Time from start of %s command "
                         "#%d to end was %d.%06d secs", read_str, do_time,
                         (int)res_tm.tv_sec, (int)res_tm.tv_usec);
@@ -844,7 +849,7 @@ int main(int argc, char * argv[])
         int fd;
         char c;
 
-        fprintf(stderr, ">> Direct IO requested but incomplete %d times\n", 
+        fprintf(stderr, ">> Direct IO requested but incomplete %d times\n",
                 dio_incomplete);
         if ((fd = open(proc_allow_dio, O_RDONLY)) >= 0) {
             if (1 == read(fd, &c, 1)) {
@@ -856,7 +861,7 @@ int main(int argc, char * argv[])
         }
     }
     if (sum_of_resids)
-        fprintf(stderr, ">> Non-zero sum of residual counts=%d\n", 
+        fprintf(stderr, ">> Non-zero sum of residual counts=%d\n",
                 sum_of_resids);
     return (ret >= 0) ? ret : SG_LIB_CAT_OTHER;
 }
